from typing import List, Sequence, Tuple
from dataclasses import asdict
from scipy.spatial import ConvexHull

import cv2
import numpy as np

from scipy.interpolate import splprep, splev
from ultralytics import YOLO

from pflows.polygons import iou_polygons
from pflows.tools.categories import remap_category_ids
from pflows.typedef import Annotation, Category, Dataset, Image


def filter_by_tag(
    dataset: Dataset, include: List[str] | None = None, exclude: List[str] | None = None
) -> Dataset:
    include = include or []
    exclude = exclude or []

    return Dataset(
        images=[
            Image(
                **{
                    **asdict(image),
                    "annotations": [
                        annotation
                        for annotation in image.annotations
                        if (
                            len(include) == 0
                            or any(tag in (annotation.tags or []) for tag in include)
                        )
                        and (
                            len(exclude) == 0
                            or all(tag not in (annotation.tags or []) for tag in exclude)
                        )
                    ],
                }
            )
            for image in dataset.images
        ],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def keep_certain_categories(dataset: Dataset, categories: List[str]) -> Dataset:
    return Dataset(
        images=[
            Image(
                **{
                    **asdict(image),
                    "annotations": [
                        annotation
                        for annotation in image.annotations
                        if annotation.category_name in categories
                    ],
                }
            )
            for image in dataset.images
        ],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def new_annotation(annotation, overwrite: dict):
    return Annotation(
        **{
            **asdict(annotation),
            **overwrite,
        }
    )


def change_all_categories(dataset: Dataset, new_category: str) -> Dataset:
    return Dataset(
        images=[
            Image(
                **{
                    **asdict(image),
                    "annotations": [
                        new_annotation(
                            annotation, {"category_name": new_category, "category_id": 0}
                        )
                        for annotation in image.annotations
                    ],
                }
            )
            for image in dataset.images
        ],
        categories=[Category(id=0, name=new_category)],
        groups=dataset.groups,
    )


def remove_annotations(
    dataset: Dataset, remove_tags: List[str], remove_categories_model: str | None = None
) -> Dataset:
    categories = dataset.categories
    if remove_categories_model:
        model = model = YOLO(remove_categories_model)
        model_categories = model.names
        if isinstance(model_categories, dict):
            model_categories = model_categories.values()
        categories = [category for category in categories if category.name not in model_categories]
    return remap_category_ids(
        Dataset(
            images=[
                Image(
                    **{
                        **asdict(image),
                        "annotations": [
                            annotation
                            for annotation in image.annotations
                            if all(tag not in (annotation.tags or []) for tag in remove_tags)
                        ],
                    }
                )
                for image in dataset.images
            ],
            categories=categories,
            groups=dataset.groups,
        )
    )


def chaikin_smooth(raw_polygon, iterations=5, tension=0.75):
    # Convert raw_polygon to [(x1, y1), (x2, y2), ...]
    polygon = [(raw_polygon[i], raw_polygon[i + 1]) for i in range(0, len(raw_polygon), 2)]

    for _ in range(iterations):
        smoothed = []
        for i in range(len(polygon)):
            x0, y0 = polygon[i]
            x1, y1 = polygon[(i + 1) % len(polygon)]

            # Create two new points using the tension parameter
            smoothed.append((tension * x0 + (1 - tension) * x1, tension * y0 + (1 - tension) * y1))
            smoothed.append(((1 - tension) * x0 + tension * x1, (1 - tension) * y0 + tension * y1))

        polygon = smoothed

    # Convert back to the format of [x1, y1, x2, y2, ...]
    return [coord for point in polygon for coord in point]


def b_spline_smoothing(polygon, smoothness=0.0005, num_points=70):
    # Convertir el polígono a un array numpy
    points = np.array(polygon).reshape(-1, 2)

    # Cerrar el polígono si no lo está
    if not np.array_equal(points[0], points[-1]):
        points = np.vstack((points, points[0]))

    # Aplicar spline paramétrico
    tck, u = splprep(points.T, u=None, s=smoothness, per=1)
    u_new = np.linspace(u.min(), u.max(), num_points)
    x_new, y_new = splev(u_new, tck, der=0)

    # Convertir de vuelta a la forma original
    smoothed_polygon = list(zip(x_new, y_new))

    # flatten
    smoothed_polygon = [coord for point in smoothed_polygon for coord in point]
    return smoothed_polygon


def smooth_segments(
    dataset: Dataset,
    technique="spline",
    smoothness: float = 0.0005,
    num_points: int = 70,
    smooth_iterations: int = 1,
    tension: float = 0.75,
) -> Dataset:
    if technique == "spline":
        smooth_polygon = b_spline_smoothing
        args = {"smoothness": smoothness, "num_points": num_points}
    elif technique == "chaikin":
        smooth_polygon = chaikin_smooth
        args = {"iterations": smooth_iterations, "tension": tension}
    else:
        raise ValueError("Invalid smoothing technique")
    return Dataset(
        images=[
            Image(
                **{
                    **asdict(image),
                    "annotations": [
                        (
                            new_annotation(
                                annotation,
                                {"segmentation": smooth_polygon(annotation.segmentation, **args)},
                            )
                            if annotation.task == "segment" and annotation.segmentation
                            else annotation
                        )
                        for annotation in image.annotations
                    ],
                }
            )
            for image in dataset.images
        ],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def smooth_image(image: Image, smoothness=0.0001, num_points=70):
    for annotation in image.annotations:
        if annotation.task == "segment":
            smoothed_polygon = smooth_polygon(annotation.segmentation, smoothness, num_points)
            annotation.segmentation = smoothed_polygon
    return image


def is_complete_annotation(
    annotation: Annotation, image_width: int, image_height: int, border_threshold=10
) -> bool:
    if not annotation or annotation.segmentation is None or annotation.bbox is None:
        return False

    if annotation.bbox:
        x1, y1, x2, y2 = annotation.bbox
        # adjusted x1, y1, x2, y2
        x1 = x1 * image_width
        y1 = y1 * image_height
        x2 = x2 * image_width
        y2 = y2 * image_height
        return (
            border_threshold < x1 < x2 < image_width - border_threshold
            and border_threshold < y1 < y2 < image_height - border_threshold
        )

    points = [
        (point[0] * image_width, point[1] * image_height)
        for point in np.array(annotation.segmentation).reshape(-1, 2)
    ]

    # Create a mask for this segment
    mask = np.zeros((image_height, image_width), dtype=np.uint8)
    cv2.fillPoly(mask, [points], 255)

    # Find contours
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    if len(contours) == 1:
        contour = contours[0]
        x, y, w, h = cv2.boundingRect(contour)

        # Check if the contour is away from the image border
        if (
            x > border_threshold
            and y > border_threshold
            and x + w < image_width - border_threshold
            and y + h < image_height - border_threshold
        ):

            # Check aspect ratio
            aspect_ratio = h / w
            if 1.5 < aspect_ratio < 5:  # Adjust these values as needed

                # Additional check: ensure the segment doesn't touch the bottom
                if np.max(points[:, 1]) < image_height - border_threshold:
                    return True
    return False


def select_image_complete_annotation(image: Image, border_tolerance: int = 10) -> Image:
    complete_annotations = [
        ann
        for ann in image.annotations
        if is_complete_annotation(ann, image.width, image.height, border_tolerance)
    ]
    return Image(
        **{
            **asdict(image),
            "annotations": complete_annotations,
        }
    )


def select_complete_annotations(dataset: Dataset, border_tolerance: int = 10) -> Dataset:
    new_images = []
    for image in dataset.images:
        new_images.append(select_image_complete_annotation(image, border_tolerance))

    return Dataset(images=new_images, categories=dataset.categories, groups=dataset.groups)


def remove_duplicate_annotations(dataset: Dataset, iou_threshold: float = 0.5) -> Dataset:
    new_images = []
    for image in dataset.images:
        new_annotations = []
        for i, annotation in enumerate(image.annotations):
            if annotation.segmentation is None:
                new_annotations.append(annotation)
                continue

            is_duplicate = False
            for other_annotation in image.annotations[i + 1 :]:
                if other_annotation.segmentation is None:
                    continue

                iou = iou_polygons(annotation.segmentation, other_annotation.segmentation)
                if iou > iou_threshold:
                    is_duplicate = True
                    break

            if not is_duplicate:
                new_annotations.append(annotation)

        new_images.append(Image(**{**asdict(image), "annotations": new_annotations}))

    return Dataset(images=new_images, categories=dataset.categories, groups=dataset.groups)


def calculate_bbox_iou(bbox1, bbox2):
    x1 = max(bbox1[0], bbox2[0])
    y1 = max(bbox1[1], bbox2[1])
    x2 = min(bbox1[2], bbox2[2])
    y2 = min(bbox1[3], bbox2[3])

    intersection = max(0, x2 - x1) * max(0, y2 - y1)
    area1 = (bbox1[2] - bbox1[0]) * (bbox1[3] - bbox1[1])
    area2 = (bbox2[2] - bbox2[0]) * (bbox2[3] - bbox2[1])

    iou = intersection / (area1 + area2 - intersection)
    return iou


def replace_category_on_overlap(
    dataset: Dataset,
    search_category: str,
    replace_categories: List[str],
    new_category: str,
    iou_threshold: float = 0.5,
) -> Dataset:
    new_categories = dataset.categories
    if new_category not in [cat.name for cat in dataset.categories]:
        new_categories = dataset.categories + [
            Category(id=len(dataset.categories), name=new_category)
        ]

    def process_image(image: Image) -> Image:
        search_annotations = [
            ann for ann in image.annotations if ann.category_name == search_category
        ]
        replace_annotations = [
            ann for ann in image.annotations if ann.category_name in replace_categories
        ]
        other_annotations = [
            ann
            for ann in image.annotations
            if ann.category_name not in [search_category] + replace_categories
        ]

        new_annotations = []
        for replace_ann in replace_annotations:
            overlap_found = False
            for search_ann in search_annotations:
                if replace_ann.segmentation and search_ann.segmentation:
                    iou = iou_polygons(replace_ann.segmentation, search_ann.segmentation)
                    if iou > iou_threshold:
                        overlap_found = True
                        break
                elif replace_ann.bbox and search_ann.bbox:
                    iou = calculate_bbox_iou(replace_ann.bbox, search_ann.bbox)
                    if iou > iou_threshold:
                        overlap_found = True
                        break

            if overlap_found:
                new_annotations.append(
                    Annotation(
                        id=replace_ann.id,
                        category_id=next(
                            cat.id for cat in new_categories if cat.name == new_category
                        ),
                        category_name=new_category,
                        center=replace_ann.center,
                        bbox=replace_ann.bbox,
                        segmentation=replace_ann.segmentation,
                        task=replace_ann.task,
                        conf=replace_ann.conf,
                        tags=replace_ann.tags,
                    )
                )
            else:
                new_annotations.append(replace_ann)

        # Add all search_category annotations and other annotations
        new_annotations.extend(search_annotations)
        new_annotations.extend(other_annotations)

        return Image(**{**image.__dict__, "annotations": new_annotations})

    new_images = [process_image(image) for image in dataset.images]

    return Dataset(images=new_images, categories=new_categories, groups=dataset.groups)


def change_annotation_from_another_annotation(
    dataset: Dataset,
    search_annotation_category_names: Sequence[str],
    replace_annotation_category_name: Sequence[str],
) -> Dataset:
    for image in dataset.images:
        tooth_numbering_annotations = [
            annotation
            for annotation in image.annotations
            if annotation.category_name in replace_annotation_category_name
        ]
        segmentation_annotations = [
            annotation
            for annotation in image.annotations
            if annotation.category_name in search_annotation_category_names
        ]

        for annotation in segmentation_annotations:
            # find closest center from tooth_numbering_annotations
            closest = None
            min_distance = float("inf")
            for tooth_numbering_annotation in tooth_numbering_annotations:
                distance = annotation.distance(tooth_numbering_annotation)
                if distance < min_distance:
                    min_distance = distance
                    closest = tooth_numbering_annotation
            if closest:
                annotation.category_id = closest.category_id
                annotation.category_name = closest.category_name

    return dataset


def polygon_to_obb(polygon):
    if polygon is None:
        return None
    if len(polygon) < 3:
        return None
    # Ensure polygon is a numpy array
    polygon = np.array(polygon).reshape(-1, 2)

    # Get the convex hull of the polygon
    hull = ConvexHull(polygon)
    hull_points = polygon[hull.vertices]

    # Function to calculate the area of a rectangle
    def rect_area(rect):
        return np.linalg.norm(rect[1] - rect[0]) * np.linalg.norm(rect[2] - rect[1])

    # Initialize variables
    min_rect = None
    min_area = float("inf")

    # Iterate through all edges of the convex hull
    for i in range(len(hull_points)):
        edge = hull_points[(i + 1) % len(hull_points)] - hull_points[i]
        edge_norm = edge / np.linalg.norm(edge)

        # Create orthogonal vector
        orth = np.array([-edge_norm[1], edge_norm[0]])

        # Project all points onto the edge and its orthogonal
        proj_edge = np.dot(hull_points - hull_points[i], edge_norm)
        proj_orth = np.dot(hull_points - hull_points[i], orth)

        # Find min and max projections
        min_edge, max_edge = np.min(proj_edge), np.max(proj_edge)
        min_orth, max_orth = np.min(proj_orth), np.max(proj_orth)

        # Calculate rectangle vertices
        rect = np.array(
            [
                hull_points[i] + min_edge * edge_norm + min_orth * orth,
                hull_points[i] + max_edge * edge_norm + min_orth * orth,
                hull_points[i] + max_edge * edge_norm + max_orth * orth,
                hull_points[i] + min_edge * edge_norm + max_orth * orth,
            ]
        )

        # Update minimum area rectangle if necessary
        area = rect_area(rect)
        if area < min_area:
            min_area = area
            min_rect = rect

    # Flatten the rectangle coordinates
    return min_rect.flatten().tolist()


def image_to_obb(image: Image) -> Image:
    return Image(
        **{
            **asdict(image),
            "annotations": [
                Annotation(
                    **{
                        **asdict(ann),
                        "bbox": polygon_to_obb(ann.segmentation),
                        "task": "obb",
                    }
                )
                for ann in image.annotations
            ],
        }
    )


def dataset_to_obb(dataset: Dataset) -> Dataset:
    return Dataset(
        images=[image_to_obb(image) for image in dataset.images],
        categories=dataset.categories,
        groups=dataset.groups,
    )



import json
import sys
import cv2
import os
import numpy as np
import yaml
import os.path


def putTextWithShadow(
    image,
    label,
    position,
    font,
    font_scale,
    text_color,
    shadow_color,
    text_thickness=1,
    shadow_thickness=1,
    shadow_offset=2,
):
    shadow_position = (position[0] + shadow_offset, position[1] + shadow_offset)

    # Draw shadow
    cv2.putText(image, label, shadow_position, font, font_scale, shadow_color, shadow_thickness)

    # Draw text
    cv2.putText(image, label, position, font, font_scale, text_color, text_thickness)


def load_original_image(image_path):
    original_image_path = image_path.replace("/yolo/train/images", "/raw/train/images")
    original_image_path = original_image_path.replace("/yolo/test/images", "/raw/test/images")
    original_image_path = original_image_path.replace("/yolo/val/images", "/raw/val/images")

    if not os.path.exists(original_image_path):
        original_image_path = image_path
    return original_image_path


def create_legend(classes, colors):
    # add ALL category
    classes = ["ALL"] + classes
    legend_image = np.zeros((len(classes) * 25, 200, 3), dtype=np.uint8)
    for i, class_name in enumerate(classes):
        if i == 0:
            color = (255, 255, 255)
        else:
            color = colors[i - 1]
        cv2.rectangle(legend_image, (0, i * 25), (25, (i + 1) * 25), color, -1)
        cv2.putText(
            legend_image,
            class_name,
            (30, i * 25 + 20),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.6,
            (255, 255, 255),
            1,
        )
    return legend_image


def load_colors(classes):
    colors = [
        (255, 0, 0),  # Red
        (0, 255, 0),  # Green
        (0, 0, 255),  # Blue
        (255, 255, 0),  # Yellow
        (255, 0, 255),  # Magenta
        (0, 255, 255),  # Cyan
        (128, 0, 0),  # Maroon
        (0, 128, 0),  # Dark Green
        (0, 0, 128),  # Navy Blue
        (128, 128, 0),  # Olive
        (128, 0, 128),  # Purple
        (0, 128, 128),  # Teal
        (255, 165, 0),  # Orange
        (255, 192, 203),  # Pink
        (165, 42, 42),  # Brown
        (240, 230, 140),  # Khaki
        (173, 216, 230),  # Light Blue
        (240, 128, 128),  # Light Coral
        (144, 238, 144),  # Light Green
        (224, 255, 255),  # Light Cyan
    ]
    colors = colors * (len(classes) // len(colors) + 1)
    return colors[: len(classes)]


def load_classes(image_path):
    yaml_dir = image_path.split("/images/")[0]
    if yaml_dir == image_path:
        yaml_dir = os.path.dirname(image_path)

    classes = []
    try:

        yaml_file = open(f"{yaml_dir}/../data.yaml", "r")
        parsed_yaml_file = yaml.load(yaml_file, Loader=yaml.FullLoader)
        classes = parsed_yaml_file["names"]
        if isinstance(classes, dict):
            classes = list(classes.values())
        return [str(cls) for cls in classes]
    except:
        try:
            classes_file = open(f"{yaml_dir}/../classes.txt", "r")
            return classes_file.read().splitlines()
        except:
            try:
                print(f"{yaml_dir}/../../classes.txt")
                classes_file = open(f"{yaml_dir}/../../classes.txt", "r")
                return classes_file.read().splitlines()
            except:
                return []


def get_label_path(image_path):
    label_path = image_path.replace(".jpg", ".txt").replace("/images/", "/labels/")
    confidence_path = image_path.replace(".jpg", ".txt").replace("/images/", "/confidences/")
    if os.path.exists(label_path):
        if confidence_path == image_path or confidence_path == label_path:
            confidence_path = None
        return label_path, confidence_path
    label_path = image_path.replace(".jpg", ".txt")
    if os.path.exists(label_path):
        return label_path, None
    raise Exception("No label path found")


def get_color(classes, colors, class_id):
    if isinstance(classes, dict):
        return colors[classes[class_id]]
    return colors[class_id]


def load_image(image_path, selected_category, legend_image, classes, colors, initial=False):
    if initial:
        print("image_path", image_path)
    original_image_path = load_original_image(image_path)

    label_path, confidence_path = get_label_path(image_path)
    info_path = image_path.replace(".jpg", ".json").replace("/images/", "/info/")
    if info_path and info_path != image_path and os.path.exists(info_path):
        info = json.loads(open(info_path, "r").read())
        print("\tsource:", info.get("path"))
    # Load the image
    image = cv2.imread(original_image_path)

    # YOLOv8 output format for a single bounding box: [class_id, center_x, center_y, width, height]
    # You'll need to replace this with your actual YOLOv8 output data

    # read yolo output open the file and split spaces in each line
    yolo_output = []
    category_ids = set()

    with open(label_path, "r") as file:
        for line in file:
            yolo_output.append([float(x) for x in line.split(" ")])
            category_ids.add(line.split(" ")[0])
    if not classes:
        classes = dict(
            [(int(category_id), index) for index, category_id in enumerate(list(category_ids))]
        )
        colors = load_colors(classes.values())

    yaml_dir = image_path.split("/images/")[0]
    if yaml_dir == image_path:
        yaml_dir = os.path.dirname(image_path)

    confidences = None
    if confidence_path and os.path.exists(confidence_path):
        confidences = [
            float(confidence or 0) for confidence in open(confidence_path, "r").read().splitlines()
        ]

    for index, box in enumerate(yolo_output):
        left = top = right = bottom = 0

        # Define the color for the bounding box based on class_id (you can customize this)
        # color = (0, 255, 0)  # Green
        # color = (255, 255, 0)  # Blue
        color = (153, 255, 255)  # Yellow
        # color = (0, 0,0) # Black

        class_id = int(box[0])
        if selected_category is not None and class_id != selected_category:
            continue
        if len(box) > 5:
            # We are going to draw the segmentation mask
            points = box[1:]
            points = np.array(points).reshape(-1, 2)
            points = points * np.array([image.shape[1], image.shape[0]])
            points = points.astype(int)
            cv2.polylines(
                image,
                [points],
                isClosed=True,
                color=get_color(classes, colors, class_id),
                thickness=2,
            )
            # get left, top, right, bottom
            left = np.min(points[:, 0])
            top = np.min(points[:, 1])
            right = np.max(points[:, 0])
            bottom = np.max(points[:, 1])

        else:
            _, center_x, center_y, width, height = box

            # Convert YOLO format to pixel coordinates
            img_height, img_width, _ = image.shape
            left = int((center_x - width / 2) * img_width)
            top = int((center_y - height / 2) * img_height)
            right = int((center_x + width / 2) * img_width)
            bottom = int((center_y + height / 2) * img_height)

            # Draw the bounding box on the image
            cv2.rectangle(image, (left, top), (right, bottom), get_color(classes, colors, class_id))

        # Optionally, you can add text to label the object with its class ID
        label = str(classes[int(class_id)])
        putTextWithShadow(
            image, label, (left, top), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color, (0, 0, 0)
        )

        if confidences is not None:
            confidence = f"{confidences[index]:.2f}"
            putTextWithShadow(
                image, confidence, (left, top + 20), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color, (0, 0, 0)
            )
    cv2.imshow("Legend", legend_image)
    cv2.imshow("YOLOv8 Output", image)


def main(target_path):
    # Check if the target path exists and if it's a folder
    if not os.path.exists(target_path):
        print(f"Invalid path: {target_path}")
        return
    if not os.path.isdir(target_path):
        images = [target_path]
    else:
        images = [
            os.path.join(target_path, image)
            for image in os.listdir(target_path)
            if image.lower().endswith((".jpg", ".png", ".jpeg"))
        ]

    classes = load_classes(target_path)
    colors = load_colors(classes)

    legend_image = create_legend(classes, colors)
    selected_category = None
    processed_image = None
    mouse_inside_legend = False

    def mouse_callback(event, x, y, flags, param):
        nonlocal selected_category, mouse_inside_legend
        nonlocal processed_image
        if event == cv2.EVENT_MOUSEMOVE:
            if x < legend_image.shape[1] and y < legend_image.shape[0]:
                mouse_inside_legend = True
                new_category = y // 25
                if new_category == 0:
                    selected_category = None
                else:
                    selected_category = new_category - 1
                processed_image = load_image(
                    images[current_image_index], selected_category, legend_image, classes, colors
                )

    legend_window_name = "Legend"
    output_window_name = "YOLOv8 Output"

    cv2.namedWindow(legend_window_name)
    cv2.namedWindow(output_window_name)

    legend_window_width = legend_image.shape[1]

    # Set the positions of the windows
    cv2.moveWindow(legend_window_name, 100, 100)
    cv2.moveWindow(output_window_name, 100 + legend_window_width + 10, 100)

    cv2.setMouseCallback("Legend", mouse_callback)

    current_image_index = 0
    while True:
        print(f"Image {current_image_index + 1}/{len(images)}")
        if processed_image is None:
            processed_image = load_image(
                images[current_image_index],
                selected_category,
                legend_image,
                classes,
                colors,
                initial=True,
            )
        key = cv2.waitKey(0)
        if key == ord("q"):
            break
        elif key == 3:  # Left arrow
            current_image_index = (current_image_index + 1) % len(images)
        elif key == 2:  # Right arrow
            current_image_index = (current_image_index - 1) % len(images)

    cv2.destroyAllWindows()


if __name__ == "__main__":
    main(sys.argv[1])



import os
import json
import argparse
import subprocess
import sys
from dataclasses import asdict, is_dataclass
from typing import Dict, Any
from dotenv import load_dotenv

from pflows.workflow import run_workflow
from compare_models import compare_models, show_model_details
from pflows.viewer.review_images import main as review_main
from train_remote import main as train_main
from run_model import run_and_compare


load_dotenv(".env")


def format_output(output_data: Dict[str, Any]) -> Dict[str, Any]:
    return {
        key: asdict(value) if is_dataclass(value) else value for key, value in output_data.items()
    }


def run_command(workflow_path: str, output_json: str, env_path: str) -> None:
    output_data: Dict[str, Dict[str, Any]] = {"job": {}}
    output_key = "job"

    try:
        if env_path:
            load_dotenv(env_path)
        output_data = run_workflow(workflow_path, store_dict=output_data, store_dict_key=output_key)
        formatted_output = format_output(output_data)
        if output_json:
            with open(output_json, "w", encoding="utf-8") as f:
                f.write(json.dumps(formatted_output, indent=4))
    except Exception as e:
        raise e


def modal_server_command() -> None:
    current_file = os.path.abspath(__file__)
    main_folder = os.path.abspath(os.path.join(os.path.dirname(current_file), "..", ".."))
    subprocess.run(["modal", "serve", "./src/modal_server.py"], cwd=main_folder)


def train_remote_command(config: str) -> None:
    train_main(config)


def review_command(filepath: str) -> None:
    review_main(filepath)


def main():
    parser = argparse.ArgumentParser(description="pflows command line tool")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Parser for the 'run' command
    run_parser = subparsers.add_parser("run", help="Run a workflow")
    run_parser.add_argument("workflow_path", type=str, help="Path to the workflow file")
    run_parser.add_argument(
        "--output_json", default=None, type=str, help="Path to the output JSON file"
    )
    run_parser.add_argument("--env", default=None, type=str, help="Path to another env")

    # Parser for the 'modal_server' command
    subparsers.add_parser("modal_server", help="Run modal server")

    # Train server
    train_parser = subparsers.add_parser("train_remote", help="Run a train process")
    train_parser.add_argument("config", type=str, help="Path to the config file")

    # Parser for the 'review' command
    review_parser = subparsers.add_parser("review", help="Run review script")
    review_parser.add_argument("filepath", type=str, help="Path to the file to review")

    # Parser for the 'compare' command
    compare_parser = subparsers.add_parser("compare", help="Compare metrics across multiple models")
    compare_parser.add_argument("folder", help="Root folder containing model metrics")
    compare_parser.add_argument(
        "name_filter", nargs="?", default=None, help="Optional name filter for models"
    )
    compare_parser.add_argument(
        "--format", choices=["markdown", "csv", "tsv"], default="markdown", help="Output format"
    )
    compare_parser.add_argument(
        "--sort", choices=["precision", "recall", "f1"], default=None, help="Sort results by metric"
    )

    # Parser for the 'detail' command
    detail_parser = subparsers.add_parser(
        "detail", help="Show detailed metrics for a specific model"
    )
    detail_parser.add_argument(
        "model_path",
        help="Path to the model file or folder containing the model's metrics.json file",
    )
    detail_parser.add_argument(
        "--format", choices=["markdown", "csv", "tsv"], default="markdown", help="Output format"
    )
    detail_parser.add_argument(
        "--sort",
        choices=["precision", "recall", "f1"],
        default=None,
        help="Sort class results by metric",
    )

    # Parser for the 'run_model' command
    run_model_parser = subparsers.add_parser("run_model", help="Run YOLO model and compare with gold standard")
    run_model_parser.add_argument("model_path", help="Path to YOLO model (.pt file)")
    run_model_parser.add_argument("dataset_path", help="Path to YOLO dataset folder")
    run_model_parser.add_argument("--groups", nargs="+", help="Groups to evaluate (e.g. train val)")
    run_model_parser.add_argument("--threshold", type=float, default=0.5, help="Confidence threshold")
    run_model_parser.add_argument("--iou-threshold", type=float, default=0.5, help="IoU threshold")
    run_model_parser.add_argument("--output", help="Path to save metrics JSON file")

    args = parser.parse_args()

    if args.command == "run":
        run_command(args.workflow_path, args.output_json, args.env)
    elif args.command == "modal_server":
        modal_server_command()
    elif args.command == "train_remote":
        train_remote_command(args.config)
    elif args.command == "review":
        review_command(args.filepath)
    elif args.command == "compare":
        result = compare_models(args.folder, args.name_filter, args.format, args.sort)
        print(result)
    elif args.command == "detail":
        result = show_model_details(args.model_path, args.format, args.sort)
        print(result)
    elif args.command == "run_model":
        metrics = run_and_compare(
            args.model_path,
            args.dataset_path,
            args.groups,
            args.threshold,
            args.iou_threshold,
            args.output,
        )
        
        # Print summary metrics
        print("\nOverall Metrics:")
        print(f"Precision: {metrics['overall']['precision']:.4f}")
        print(f"Recall: {metrics['overall']['recall']:.4f}")
        print(f"F1 Score: {metrics['overall']['f1_score']:.4f}")

        print("\nMetrics by Category:")
        for category, cat_metrics in metrics["categories"].items():
            print(f"\n{category}:")
            print(f"  Precision: {cat_metrics['precision']:.4f}")
            print(f"  Recall: {cat_metrics['recall']:.4f}")
            print(f"  F1 Score: {cat_metrics['f1_score']:.4f}")
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import os
import re
import json
import inspect

from typing import Sequence, Tuple, Dict, Any, Callable, List, cast
from pflows.typedef import Dataset, Task


def load_function_args(function: Callable[..., Any]) -> Dict[str, Dict[str, str | bool]]:
    # Get the function signature
    signature = inspect.signature(function)
    # Create a dictionary to store the parameter information
    param_info = {}

    # Iterate over the parameters and populate the dictionary
    for param in signature.parameters.values():
        param_info[param.name] = {
            "type": param.annotation,
            "required": param.default == inspect.Parameter.empty,
        }
    return param_info


def load_function(task_name: str) -> Tuple[Any, Dict[str, Dict[str, str | bool]]]:
    try:
        task_file, task_function_name = task_name.split(".")
        task_module = __import__(f"pflows.tools.{task_file}", fromlist=[task_function_name])
        task_function = getattr(task_module, task_function_name)
        params = load_function_args(task_function)
    except Exception as exc:
        print(f"Error loading task: {task_name}", exc)
        raise ValueError(f"The task '{task_name}' is not a valid task.") from exc

    return task_function, params


def replace_variables(text: str, current_dir: str | None = None) -> str:
    # we are going to search for {{variable}} and replace it with the value of the variable
    # we are going to use a regular expression to find all the variables
    matches = re.findall(r"\{\{([a-zA-Z0-9_]+)\}\}", text)
    for match in matches:
        value = os.getenv(match)
        if value is None or value == "":
            if match == "CURRENT_DIR" and current_dir is not None:
                value = current_dir
            else:
                found_base_folder = os.getenv("BASE_FOLDER")
                if match == "PERSISTED_FOLDER" and found_base_folder is not None:
                    value = found_base_folder
                else:
                    print(f"Variable {match} not found in environment variables.")
                    continue
        text = text.replace(f"{{{{{match}}}}}", value)
    return text


def prepare_workflow(
    raw_workflow: Sequence[Dict[str, Any]], workflow_dir: str | None
) -> Sequence[Dict[str, Any]]:
    non_set_env_tasks_found = False
    workflow_tasks = []
    for task in raw_workflow:
        if task["task"] == "set_env_var":
            if non_set_env_tasks_found:
                raise ValueError("set_env_var tasks must be the first tasks in the workflow.")
            os.environ[task["name"]] = task["value"]
            continue
        workflow_tasks.append(task)
        non_set_env_tasks_found = True

    # Keep in a variable all the functions that are going to be used in the workflow
    workflow_functions = {}
    for index, task in enumerate(workflow_tasks):
        if task.get("function"):
            workflow_functions[f"task_{index}"] = task["function"]
            task["function"] = "run_function_path"

    workflow_text = json.dumps(workflow_tasks)
    workflow_text = replace_variables(workflow_text, workflow_dir)
    workflow = cast(Sequence[Dict[str, Any]], json.loads(workflow_text))
    for index, task in enumerate(workflow):
        if task.get("function"):
            task["function"] = workflow_functions[f"task_{index}"]
    return workflow


# pylint: disable=too-many-locals,too-many-branches
def read_workflow(
    workflow_path: str | None = None, raw_workflow: Sequence[Dict[str, Any]] | None = None
) -> Tuple[Sequence[Task], Dict[str, Any]]:
    if workflow_path is None and raw_workflow is None:
        raise ValueError("You must provide a workflow path or a workflow.")
    workflow_dir = None
    if workflow_path is not None:
        raw_workflow, workflow_dir = read_local_workflow(workflow_path)
    if raw_workflow is None:
        raise ValueError("You must provide a workflow path or a workflow.")

    workflow = prepare_workflow(raw_workflow, workflow_dir)

    workflow_data = {"dataset": Dataset(images=[], categories=[], groups=[])}
    workflow_reviewed_tasks: List[Task] = []
    for index, raw_task in enumerate(workflow):
        if "task" not in raw_task:
            raise ValueError(f"The 'task' key is missing in one of the task {index +1}.")
        task_name = raw_task["task"]
        task_args = raw_task.copy()
        del task_args["task"]

        task_function, params = load_function(task_name)
        skip_task = False
        task_id = None
        # check if all required parameters are present
        for param, info in params.items():
            if info["required"] and param not in task_args:
                possible_relatives = [
                    task_param
                    for task_param in task_args
                    if task_param.endswith("_relative")
                    and re.sub(r"_relative$", "", task_param) == param
                ]
                if len(possible_relatives) > 0 and workflow_dir is not None:
                    relative_param = possible_relatives[0]
                    task_args[param] = os.path.abspath(
                        os.path.join(workflow_dir, task_args[relative_param])
                    )
                    del task_args[relative_param]
                    continue
                if param in workflow_data:
                    task_args[param] = "__workflow_parameter__"
                    continue
                raise ValueError(
                    f"The parameter '{param}' is required for task {index +1}: {raw_task['task']}."
                )
        if "skip" in task_args:
            skip_task = True
            del task_args["skip"]
        if "id" in task_args:
            task_id = task_args["id"]
            del task_args["id"]

        # check if there are any extra parameters
        for param in task_args:
            if param not in params:
                if param in workflow_data:
                    task_args[param] = workflow_data[param]
                    continue
                raise ValueError(f"The parameter '{param}' is not valid for task {index +1}.")
        task = Task(
            task=task_name, function=task_function, params=task_args, skip=skip_task, id=task_id
        )
        workflow_reviewed_tasks.append(task)

    return workflow_reviewed_tasks, workflow_data


def read_local_workflow(workflow_path: str) -> Tuple[Sequence[Dict[str, Any]], str]:
    # Load the workflow and check is a valid JSON
    if not os.path.exists(workflow_path):
        raise FileNotFoundError("The workflow file does not exist.")
    workflow_dir = os.path.abspath(os.path.dirname(workflow_path))

    with open(workflow_path, "r", encoding="utf-8") as f:
        try:
            workflow_text = f.read()
            workflow = json.loads(workflow_text)
        except json.JSONDecodeError as exc:
            raise ValueError("The workflow file is not a valid JSON file.") from exc
    return workflow, workflow_dir


def internal_run_workflow(
    workflow: Sequence[Task],
    workflow_data: Dict[str, Any],
    store_dict: Dict[str, Any] | None = None,
    store_dict_key: str | None = None,
    id_output_data: Dict[str, Any] = {},
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    for task in workflow:
        print("")
        print("-" * 20, task.task, "-" * 20)
        if task.skip:
            print("Skipping task.")
            continue
        params = {
            param: (
                task.params[param]
                if task.params[param] != "__workflow_parameter__"
                else workflow_data[param]
            )
            for param in task.params
        }
        result = task.function(**params)
        if result is not None and isinstance(result, dict):
            workflow_data.update(result)
            # get result without the dataset key
            if task.id is not None:
                id_output_data[task.id] = {
                    key: value for key, value in result.items() if key != "dataset"
                }
                if store_dict is not None and store_dict_key is not None and result != {}:
                    store_dict[store_dict_key] = {
                        **store_dict[store_dict_key],
                        **id_output_data,
                    }
        if result is not None and isinstance(result, Dataset):
            workflow_data["dataset"] = result
    return workflow_data, id_output_data


def run_workflow(
    workflow_path: str | None = None,
    raw_workflow: Sequence[Dict[str, Any]] | None = None,
    store_dict: Dict[str, Any] | None = None,
    store_dict_key: str | None = None,
) -> Dict[str, Any]:
    workflow, workflow_data = read_workflow(workflow_path, raw_workflow)
    id_output_data = {}
    try:
        workflow_data, id_output_data = internal_run_workflow(
            workflow, workflow_data, store_dict, store_dict_key, id_output_data
        )

    except SystemExit:
        pass
    if store_dict is not None and store_dict_key is not None:
        store_dict[store_dict_key] = {**store_dict[store_dict_key], **id_output_data}
        return cast(Dict[str, Any], store_dict[store_dict_key])
    return id_output_data


def run_workflow_dataset(
    workflow_path: str | None = None,
    raw_workflow: Sequence[Dict[str, Any]] | None = None,
    store_dict: Dict[str, Any] | None = None,
    store_dict_key: str | None = None,
    dataset: Dataset | None = None,
) -> Dataset:
    workflow, workflow_data = read_workflow(workflow_path, raw_workflow)
    if dataset is not None:
        workflow_data["dataset"] = dataset
    try:
        workflow_data, _ = internal_run_workflow(
            workflow, workflow_data, store_dict, store_dict_key
        )
    except SystemExit:
        pass
    return Dataset.from_dict(workflow_data["dataset"])
